﻿using System;
using System.Collections.Generic;
using UnityEngine;

//Clase encargada de la gestion de las medidas recibidas del lidar
public class LidarMeasure
{
    public static readonly int measureByteSize = 6; //6 bytes (1 float + 1 uint16)
    public float lidarAngle { get; }
    public UInt16 lidarDistance { get; }
    public UInt16 servoAngle { get; }

    public LidarMeasure(float lidarAngle, UInt16 lidarDistance, UInt16 servoAngle)
    {
        this.lidarAngle = lidarAngle;
        this.lidarDistance = lidarDistance;
        this.servoAngle = servoAngle;
    }

    //Verifica los rangos utiles de las medidas
    public bool lidarMeasureOk()
    {
        if ( ((lidarAngle >= 0.001f && lidarAngle <= 205f) ||
            (lidarAngle >= 335f && lidarAngle <= 360f)) &&
            (lidarDistance >= 150 && lidarDistance <= 12000) &&
            (servoAngle >= 0 && servoAngle <= 360))
            return true;
        Debug.Log("Punto malo");
        return false;
    }

    /*Extrae la carga util de las tramas API
     * @param frame trama recibida desde Arduino 
     * @servoAngle posicion en la que se encuentra el servo
     * @return lista de la medidas contenidas en la carga util
     */
    public static LidarMeasure[] extractMeasuresFromFrame(byte[] frame, UInt16 servoAngle)
    {
        float lidarAngle;
        UInt16 lidarDistance;
        LidarMeasure[] lidarMeasures = new LidarMeasure[frame.Length / measureByteSize]; 

        for (int i = 0; i < frame.Length; i += measureByteSize)
        {
            lidarAngle = BitConverter.ToSingle(frame, i);
            lidarDistance = BitConverter.ToUInt16(frame, i + 4);
            lidarMeasures[i / measureByteSize] = new LidarMeasure(lidarAngle, lidarDistance, servoAngle);
        }
        return lidarMeasures;
    }

    /*Transforma una lista de medidas en una lista de coordenas cartesianas
     * @lidarMeasures lista de medidas a convertir
     * @carX posicion X del vehiculo en Unity (con respecto al punto inicial)
     * @carY posicion Y del vehiculo en Unity (con respecto al punto inicial)
     * @carZ posicion Z del vehiculo en Unity (con respecto al punto inicial)
     * @carAngle angulo al que se encuaentra el vehiculo en Unity (con respecto a +Z)
     * @return lista de puntos (coordenas cartesianas)
     */
    public static List<LidarPoint> transformMeasuresToPoints(LidarMeasure[] lidarMeasures, float carX, float carY, float carZ, float carAngle)
    {
        List<LidarPoint> lidarPoints = new List<LidarPoint>();
        foreach (LidarMeasure lm in lidarMeasures)
        {
            if (lm.lidarMeasureOk())
                lidarPoints.Add(lm.convertTolidarpoint(carX, carY, carZ, carAngle));
        }
        return lidarPoints;
    }

    /*Transforma una medida en coordenas cartesianas
     * @carX posicion X del vehiculo en Unity (con respecto al punto inicial)
     * @carY posicion Y del vehiculo en Unity (con respecto al punto inicial)
     * @carZ posicion Z del vehiculo en Unity (con respecto al punto inicial)
     * @carAngle angulo al que se encuaentra el vehiculo en Unity (con respecto a +Z)
     * @return punto (coordenas cartesianas)
     */
    private LidarPoint convertTolidarpoint(float carX, float carY, float carZ, float carAngle)
    {
        float x, y, z, h;
        float xOrigen, yOrigen, zOrigen;

        y = lidarDistance * Mathf.Sin(lidarAngle * Mathf.Deg2Rad);
        h = lidarDistance * Mathf.Cos(lidarAngle * Mathf.Deg2Rad);
        z = h * Mathf.Cos(servoAngle * Mathf.Deg2Rad);
        x = h * Mathf.Sin(servoAngle * Mathf.Deg2Rad);

        zOrigen = z * Mathf.Cos((carAngle - 90) * Mathf.Deg2Rad) - x * Mathf.Sin((carAngle - 90) * Mathf.Deg2Rad);
        xOrigen = x * Mathf.Cos((carAngle - 90) * Mathf.Deg2Rad) + z * Mathf.Sin((carAngle - 90) * Mathf.Deg2Rad);


        xOrigen = xOrigen + carX;
        yOrigen = y + carY + 200;
        zOrigen = zOrigen + carZ;
        return new LidarPoint(xOrigen, yOrigen, zOrigen);
    }

}
